"""Helpers to manipulate class arguments."""

from inspect import getclasstree, getfullargspec
from typing import Iterable, List, Set, Union

__all__ = ['get_init_args']


def get_all_subclasses(cls: type) -> List[type]:
    all_subclasses: List[type] = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


def get_init_args(cls: type) -> Set[str]:
    # Get the list of classes that the original_field herits from
    original_tree_class = getclasstree([cls], unique=True)
    list_classes = _flatten_list_of_types(original_tree_class)

    # From these classes, get the list of args we can use when creating the instances
    list_args: Set[str] = set()
    for cl in list_classes:
        # `getfullargspec` allows to get the arguments of a specific function - here we use
        # it on the __init__ function that initializes instances of a class
        argspec = getfullargspec(cl.__init__)  # noqa: WPS609
        list_args.update(argspec.args)
        list_args.update(argspec.kwonlyargs)

    # Remove special args
    list_args.remove('self')  # Represents the instance of the class - can't be used to create a new instance
    return list_args


NestedClasses = Iterable[Union[type, 'NestedClasses']]


def _flatten_list_of_types(list_to_flatten: NestedClasses) -> Set[type]:
    """Flatten a list of types with nested lists and tuples.

    When using the `inspect` module `getclasstree`, it returns a list of types in a hierarchical order.
    We only need the list of unique types, without any nesting.

    Examples:
    list_to_flatten = [
        (<class 'playhouse.postgres_ext.IndexedFieldMixin'>, (<class 'object'>,)),
        [
            (
                <class 'playhouse.postgres_ext.BinaryJSONField'>,
                (
                    <class 'playhouse.postgres_ext.IndexedFieldMixin'>, <class 'playhouse.postgres_ext.JSONField'>
                )
            )
        ]
    ]
    flattened_list = [
        <class 'playhouse.postgres_ext.IndexedFieldMixin'>,
        <class 'object'>,
        <class 'playhouse.postgres_ext.JSONField'>,
        <class 'playhouse.postgres_ext.BinaryJSONField'>
    ]

    Args:
        list_to_flatten (NestedClasses): The list to be flattened

    Returns:
        Set[type]: The set of classes.
    """
    ancestor_set: Set[type] = set()
    for el in list_to_flatten:
        if isinstance(el, type):
            # If the element is a type, then we want to keep it.
            ancestor_set.add(el)
        else:
            # But if we can iterate on it (list, tuple, ..), then we need to go deeper to get the types.
            sublist = _flatten_list_of_types(el)
            ancestor_set.update(sublist)
    return ancestor_set



"""Config helper."""

import base64
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Mapping, Optional, Union, cast

import toml
from dotenv import load_dotenv

from outcome.utils import env

ValidPath = Union[str, Path]
ValidConfigType = Union[str, int, float, bool, List['ValidConfigType']]
ConfigDict = Mapping[str, ValidConfigType]


class Sentinel:
    ...


NO_DEFAULT = Sentinel()


class ConfigBackend(ABC):
    @abstractmethod
    def get(self, key: str) -> ValidConfigType:  # pragma: no cover
        ...

    @abstractmethod
    def __contains__(self, key: str) -> bool:  # pragma: no cover
        ...


base_64_protocol = 'base64://'  # noqa: WPS114


class EnvBackend(ConfigBackend):
    def __init__(self):
        dotenv_file = Path.cwd() / '.env'
        load_dotenv(dotenv_path=dotenv_file, verbose=env.is_dev())

    def get(self, key: str) -> ValidConfigType:
        value = os.environ[key]

        # We can define protocols as suffix of the env variable.
        # For instance, if the env variable starts with `base64://` we will decode the rest of the string
        # with b64decode.
        #
        # This can be useful to store RSA keys as env variable since they need to be encoded as base64 strings.
        # This can also be used to specify a path to sensitive data stored in a remote database, and the get function
        # will be able to fetch the corresponding data depending on the specified protocol.
        if value.startswith(base_64_protocol):
            return base64.b64decode(value[len(base_64_protocol) :]).decode('utf-8')

        return value

    def __contains__(self, key: str) -> bool:
        return key in os.environ


class DefaultBackend(ConfigBackend):

    defaults: Mapping[str, ValidConfigType]

    def __init__(self, defaults: Mapping[str, ValidConfigType]):
        self.defaults = defaults

    def get(self, key: str) -> ValidConfigType:
        return self.defaults[key]

    def __contains__(self, key: str) -> bool:
        return key in self.defaults


class TomlBackend(ConfigBackend):

    path: Optional[ValidPath]
    config: Optional[ConfigDict]
    aliases: Optional[Mapping[str, str]]

    def __init__(self, path: Optional[ValidPath] = None, aliases: Optional[Mapping[str, str]] = None):
        self.path = path
        self.aliases = aliases
        self.config = None

    def get(self, key: str) -> ValidConfigType:
        if not self.config:
            self.load_config()
        assert self.config is not None
        return self.config[key]

    def __contains__(self, key: str) -> bool:
        if not self.config:
            self.load_config()
        return key in cast(Mapping[str, ValidConfigType], self.config)

    def load_config(self) -> None:
        assert self.path
        self.config = self.get_config(self.path, self.aliases)

    @classmethod
    def get_config(  # noqa: WPS615
        cls,
        path: ValidPath,
        aliases: Optional[Mapping[str, str]] = None,
    ) -> Mapping[str, ValidConfigType]:
        config = toml.load(path)
        config_flattened = dict(cls.flatten_keys(config))

        if aliases:
            for original, alias in aliases.items():
                config_flattened[alias.upper()] = config_flattened.pop(original.upper())

        return config_flattened

    @classmethod
    def flatten_keys(cls, value: object, key: Optional[str] = None) -> Mapping[str, ValidConfigType]:
        if not isinstance(value, dict):
            if not key:
                raise Exception('Value cannot be a non-dict without a key')

            assert isinstance(value, (str, int, float, bool, list))
            return {key.upper(): cast(ValidConfigType, value)}

        flattened: Mapping[str, ValidConfigType] = {}
        value = cast(Mapping[str, object], value)

        for k, v in value.items():
            prefix = (f'{key}_' if key else '').upper()
            flattened.update({f'{prefix}{skey}': sval for skey, sval in cls.flatten_keys(v, k).items()})

        return flattened


class Config:  # pragma: only-covered-in-unit-tests
    """This class helps with retrieving config values from a project environment.

    You can provide a path to a TOML file, typically the pyproject.toml, or just
    let the class try to extract the values from environment variables.

    Environment variables will always take precedence over the values found in the file.

    The keys from the TOML file will be flattened and transformed to uppercase, following
    environment variable conventions.

    For example:

    ```toml
    [app]
    port = 80
    ```

    Will be transformed into

    ```py
    {
        'APP_PORT': 80
    }
    ```

    If needed you can modify Config backends to change the order of priority, or to add your own backends.
    You can use add_backend method to add your own backend at the desired priority.
    If you wish to add your own backend, it needs to inherit from ConfigBackend abstract class.
    """

    backends: List[ConfigBackend]
    default_backend: DefaultBackend

    def __init__(
        self,
        path: Optional[ValidPath] = None,
        aliases: Optional[Mapping[str, str]] = None,
        defaults: Optional[Mapping[str, ValidConfigType]] = None,
    ) -> None:
        """Initialize the class with an optional config file and set of aliases.

        The aliases dict will rewrite config keys from key to value:

        ```
        aliases = {'ORIGINAL_KEY': 'NEW_KEY'}
        config = Config('some_file.toml', aliases)

        config.get('ORIGINAL_KEY')  # -> raises KeyError
        config.get('NEW_KEY')  # -> returns value
        ```

        The defaults dict is the final fallback.

        Arguments:
            path (ValidPath, optional): The path to the config file.
            aliases (Mapping[str, str], optional): The aliasing dict.
            defaults (Mapping[str, ValidConfigType], optional): A dict of hardcoded values
        """
        self.backends = [EnvBackend()]
        if path:
            self.backends.append(TomlBackend(path=path, aliases=aliases))

        self.default_backend = DefaultBackend(defaults or {})

    def get(self, key: str, default: Union[Optional[ValidConfigType], Sentinel] = NO_DEFAULT) -> Optional[ValidConfigType]:

        for backend in self.backends:
            if key in backend:
                return backend.get(key)

        try:
            return self.default_backend.get(key)
        except KeyError as exc:
            if not isinstance(default, Sentinel):
                return default
            raise exc

    def add_backend(self, backend: ConfigBackend, priority: int):
        self.backends.insert(priority, backend)


class AppConfig(Config):
    # This class wraps Config in order to modify return value of APP_NAME
    #
    # Indeed, in most pyprojecttoml files, the application name is of the form `myproject-api-app`,
    # which maps to the name of its Github repository or Docker image.
    #
    # However, in our application environment, we only want to keep `myproject`.
    # This allows for instance to use APP_NAME to fetch the right service of OTCServices.

    def get(self, key: str, default: Union[Optional[ValidConfigType], Sentinel] = NO_DEFAULT) -> Optional[ValidConfigType]:
        key_value = super().get(key, default)
        if key == 'APP_NAME':
            if key_value is default and not isinstance(default, Sentinel):
                return default
            assert isinstance(key_value, str)
            return key_value.split('-')[0]
        return key_value



"""A decorator to simplify intercepting an exception and re-raising it as another.

Example:
    You can simplify the following code:
    ```
    def some_function():
        try:
            return some_other_function()
        exception MyException as ex:
            raise MyOtherException(ex)
    ```
    It becomes
    ```
    @reraise_as(MyException, MyOtherException)
    def some_function():
        return some_other_function()
    ````
"""

import asyncio
from contextlib import contextmanager
from functools import wraps
from typing import Any, Callable, Iterator, Protocol, Type, TypeVar, cast

SE = TypeVar('SE', bound=Exception)
TE = TypeVar('TE', bound=Exception)

Mapper = Callable[[Exception, Type[TE]], TE]


F = TypeVar('F', bound=Callable[..., Any])


class Decorator(Protocol):  # pragma: no cover
    def __call__(self, fn: F) -> F:
        ...


def default_mapper(ex: Exception, target_exc: Type[TE]) -> TE:
    """Default mapper that doesn't attempt to preserve any information. Just instantiates a TargetExClass.

    Args:
        ex (Exception): The source exception.
        target_exc (TE): The target Exception class.

    Returns:
        TE: An instance of the provided tartget Exception class.
    """
    return target_exc()


# Actually performs the catching and remapping, shouldn't be called directly
@contextmanager
def remap_exception(
    source_exc: Type[Exception],
    target_exc: Type[TE],
    mapper: Mapper[TE],
) -> Iterator[None]:  # pragma: no-cover-for-integration
    try:
        yield
    except source_exc as ex:
        raise mapper(ex, target_exc)


def reraise_as(  # noqa: WPS212
    # TODO: remove type ignore, pending: https://github.com/microsoft/pyright/discussions/1799
    source_exc: Type[SE],
    target_exc: Type[TE],
    mapper: Mapper[TE] = default_mapper,  # type: ignore
) -> Decorator:  # pragma: no-cover-for-integration
    """A decorator function that intercepts raised Exceptions and maps them to a different Exception class.

    Args:
        source_exc (SourceExClass): The Exception class to intercept.
        target_exc (TargetExClass): The Exception class to raise.
        mapper (MapperFn): The mapping function that determines how to build a TargetEx from the SourceEx.

    Returns:
        Decorator
    """

    def reraise_as_decorator(fn: F) -> F:
        # If we're dealing with an async function, we need return an
        # async function
        if asyncio.iscoroutinefunction(fn):
            # Using @wraps ensures we keep the identity of the wrapped function
            @wraps(fn)
            async def async_wrapped(*args: Any, **kwargs: Any) -> Any:
                with remap_exception(source_exc, target_exc, mapper):
                    return await fn(*args, **kwargs)

            return cast(F, async_wrapped)

        # Using @wraps ensures we keep the identity of the wrapped function
        @wraps(fn)
        def sync_wrapped(*args: Any, **kwargs: Any) -> Any:
            with remap_exception(source_exc, target_exc, mapper):
                return fn(*args, **kwargs)

        return cast(F, sync_wrapped)

    return reraise_as_decorator



"""Utils CLI."""

import click

from outcome.utils.feature_set import FeatureSetModule
from outcome.utils.loader import load_obj


@click.group()
def main():
    ...


@main.command(help='List available feature flags')
@click.option('--feature-set', default='outcome.utils:feature_set')
def features(feature_set: str):
    fs = load_obj(feature_set, default_obj='feature_set', t=FeatureSetModule)
    fs.display_features()
